/*! ********************************************
 * Main.java
 * Copyright 2022 masahiro takamoto (MotChan)
 * Distributed under the MIT license. 
 * see -> license.txt
 ******************************************** */
import java.nio.ByteOrder;
import javax.sound.sampled.AudioFormat;
import javax.sound.sampled.AudioSystem;
import javax.sound.sampled.DataLine;
import javax.sound.sampled.LineUnavailableException;
import javax.sound.sampled.SourceDataLine;

import sora.graybox.GrayBox;
	
public class Main {
	public static final int AUDIO_CHANNELS = 2;  // オーディオチャンネル数
	public static final float SAMPLE_RATE = 44100.0f;  // サンプリングレート
	public static final int BLOCK_SIZE = 2048;  // ブロックサイズ(バッファサイズ)
	public static float[][] audioBuffer;  // オーディオバッファ
	public static GrayBox grayBox;  // GrayBoxシステム基幹オブジェクト(モジュラーホスト)
	public static NoiseGenerator noiseGen;  // ノイズジェネレーター(モジュール)
	
	// ストリーミング処理 -->
	private static void _streaming() {
		try {
		// オーディオライン準備 -->
			int _byteArraySize = BLOCK_SIZE * 2 * 2;
			byte[] _bytes = new byte[_byteArraySize];
			boolean _bigEndian = ByteOrder.nativeOrder() == ByteOrder.BIG_ENDIAN;
			AudioFormat _audioFormat = new AudioFormat( AudioFormat.Encoding.PCM_SIGNED, SAMPLE_RATE, Byte.SIZE * 2, 2, 4, SAMPLE_RATE, _bigEndian);    
			DataLine.Info _dataLineInfo = new DataLine.Info(SourceDataLine.class, _audioFormat);
			SourceDataLine _souceDataLine = (SourceDataLine) AudioSystem.getLine(_dataLineInfo);
			_souceDataLine.open(_audioFormat);
			_souceDataLine.start();
		// 音声処理 -->
			int _counter = 0;
			while (_counter++ < 1000) {
			// オーディオバッファのクリア -->
				float[] _outL = audioBuffer[0], // チャンネル L
						_outR = audioBuffer[1]; // チャンネル R
				for (int i = 0; i < BLOCK_SIZE; i++) {
					_outL[i] = 0.0f;
					_outR[i] = 0.0f;
				}
			// GrayBoxによる音声処理 -->
				// GrayBox.process()メソッドを使用して音声処理を行います。
				// 第1引数 : 書き込みを行うオーディオバッファ
				// 第2引数 : 書き込み開始フレーム
				// 第3引数 : 書き込みフレーム数
				// ・今回はオーディオバッファ全体を使うので、開始フレームが0、書き込みフレーム数が4096
				// ・スロットナンバー0から順にモジュール側のIGrayBoxModule.process()メソッドが呼ばれます。
				grayBox.process(audioBuffer, 0, BLOCK_SIZE);
			// オーディオバッファからオーディオストリームへ書き込み -->
				// float[] -> byte[]
				if (_bigEndian) {
					for (int i = 0, _p = 0; i < BLOCK_SIZE; i++) {
						int _ival = (int)(_outL[i] * 32767.0f);
						_bytes[_p++] = (byte)((_ival >>> 8) & 255);
						_bytes[_p++] = (byte)(_ival & 255);
						_ival = (int)(_outL[i] * 32767.0f);
						_bytes[_p++] = (byte)((_ival >>> 8) & 255);
						_bytes[_p++] = (byte)(_ival & 255);
					}
				}
				else {
					for (int i = 0, _p = 0; i < BLOCK_SIZE; i++) {
						int _ival = (int)(_outL[i] * 32767.0f);
						_bytes[_p++] = (byte)(_ival & 255);
						_bytes[_p++] = (byte)((_ival >>> 8) & 255);
						_ival = (int)(_outL[i] * 32767.0f);
						_bytes[_p++] = (byte)(_ival & 255);
						_bytes[_p++] = (byte)((_ival >>> 8) & 255);
					}
				}
				_souceDataLine.write(_bytes, 0, _byteArraySize);
			}
		// オーディオライン閉鎖 -->
			_souceDataLine.drain();
			_souceDataLine.close();
		}
		catch (LineUnavailableException e) { e.printStackTrace(); }
	}
		
	public static void main(String[] args) throws Exception {
	// オーディオバッファ作成 -->
		audioBuffer = new float[2][];
		audioBuffer[0] = new float[BLOCK_SIZE]; // チャンネル L
		audioBuffer[1] = new float[BLOCK_SIZE]; // チャンネル R
	// GrayBoxの構築 -->
		// 第1引数 : ブロックサイズ         ( 1~ )
		// 第2引数 : サンプリングレート[Hz] ( 1~ )
		// 第3引数 : オーディオチャンネル数 ( 2~ )
		// 第4引数 : モジュールスロット数   ( 1~ ) 今回はモジュールが1つなので１
		// ・初期値は必須パラメーターですが、構築時に確定できない場合は暫定値で構いません。
		//   その場合はストリーミングを開始する前に確定させてください。
		//   関連メソッド
		//   ブロックサイズ -> GrayBox.setProcessBlockSize()
		//   サンプリングレート -> GrayBox.setSampleRate()
		//   チャンネル数 -> GrayBox.setAudioChannels()
		//   スロット数 -> GrayBox.setModuleSlots()
		grayBox = new GrayBox(BLOCK_SIZE, SAMPLE_RATE, AUDIO_CHANNELS, 1);
	// 各GrayBoxモジュールの構築 -->
		// ノイズ音を出力し続けるモジュール
		noiseGen = new NoiseGenerator();
	// モジュールの接続 -->
		// GrayBox.attachModule()メソッドを使用し、ホストへモジュールを接続します。
		// 第1引数 : 接続するスロットのインデックス
		// 第2引数 : 接続するモジュールオブジェクト
		// 接続されたモジュールは接続スロットの先頭から順に処理が行われます。
		grayBox.attachModule(0, noiseGen);
	// ストリーミングの開始 -->
		_streaming();
	}
}